/*
Benchmark OpenCL
This file is in the public domain
*/

#include <iostream>
#include <iomanip>
#include <ctime>
#include <string>
#include <exception>

#include "blowfish.h"
#include "cast256.h"
#include "cast5.h"
#include "cscipher.h"
#include "des.h"
#include "gost.h"
#include "idea.h"
#include "lubyrack.h"
#include "misty1.h"
#include "rc2.h"
#include "rc5.h"
#include "rc6.h"
#include "rijndael.h"
#include "safer_sk.h"
#include "serpent.h"
#include "shark.h"
#include "skipjack.h"
#include "square.h"
#include "tea.h"
#include "threeway.h"
#include "twofish.h"
#include "xtea.h"

#include "arc4.h"
#include "isaac.h"
#include "seal.h"

#include "randpool.h"
#include "x917.h"

#include "checksum.h"
#include "haval.h"
#include "md2.h"
#include "md4.h"
#include "md5.h"
#include "rmd128.h"
#include "rmd160.h"
#include "sha1.h"
#include "sha256.h"
#include "sha512.h"
#include "tiger.h"

#include "emac.h"
#include "md5mac.h"
using namespace OpenCL_types;

extern bool produce_html;
bool print_html = false;

void benchmark_block_ciphers(u32bit);
void benchmark_stream_ciphers(u32bit);
void benchmark_rngs(u32bit);
void benchmark_hash_functions(u32bit);
void benchmark_macs(u32bit);

int benchmark(u32bit seconds)
   {
   if(produce_html) print_html = true;
   try {
      if(print_html)
         {
         std::cout << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD "
                   << "HTML 4.0 Transitional//EN\">\n"
                   << "<HTML>\n\n"
                   << "<TITLE>OpenCL Benchmarks</TITLE>\n\n"
                   << "<BODY>\n\n"
                   << "<P><TABLE BORDER CELLSPACING=1>\n"
                   << "<THEAD>\n"
                   << "<TR><TH>Algorithm                <TH>Kbytes / second\n"
                   << "<TBODY>\n";
         }
      benchmark_block_ciphers(seconds);
      benchmark_stream_ciphers(seconds);
      benchmark_rngs(seconds);
      benchmark_hash_functions(seconds);
      benchmark_macs(seconds);
      if(print_html)
         std::cout << "</TABLE>\n\n</BODY></HTML>\n";
      }
   catch(OpenCL::Exception& e)
      {
      std::cout << "OpenCL exception caught: " << e.what() << std::endl;
      return 1;
      }
   catch(std::exception& e)
      {
      std::cout << "Standard library exception caught: " << e.what()
                << std::endl;
      return 1;
      }
   catch(...)
      {
      std::cout << "Unknown exception caught." << std::endl;
      return 1;
      }
   return 0;
   }

const byte* key = (byte*)"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                         "abcdefghijklmnopqrstuvwxyz"
                         "01234567890123456789012345"
                         "00000000000000000000000000"
                         "00000000000000000000000000";

#define FACTOR (2.0 / 3.0)

void print_result(const std::string algorithm, u32bit iterations,
                  u32bit buffersize, u32bit clocks_used)
{
   double bytes_per_sec = ((double)iterations * buffersize) /
                          ((double)clocks_used / CLOCKS_PER_SEC);
   u32bit kbytes_per_sec = (u32bit)(bytes_per_sec / 1024.0);

   if(print_html)
      {
      std::cout << "<TR><TH>" << algorithm;
      for(u32bit j = 0; j != 25 - algorithm.length(); j++)
         std::cout << ' ';
      std::cout << "<TH>" << std::setw(6) << kbytes_per_sec << std::endl;
      }
   else
      {
      std::cout << algorithm << ": ";
      for(u32bit j = 0; j != 25 - algorithm.length(); j++)
         std::cout << ' ';
      std::cout << std::setw(6) << kbytes_per_sec << " kbytes/sec"
                << std::endl;
     }
}

template<typename B>
void bench_bcipher(const std::string& name, u32bit seconds)
   {
   static const u32bit BUFFERSIZE = B::BLOCKSIZE;
   B cipher;
   cipher.set_key(key);
   byte buf[BUFFERSIZE] = { 0 };
   u32bit iterations = 1, j = 0;
   u32bit start = std::clock(), clocks_used = 0;
   while(clocks_used < FACTOR * seconds * CLOCKS_PER_SEC)
      {
      iterations *= 2;
      for(; j != iterations; j++)
         cipher.encrypt(buf);
      clocks_used = std::clock() - start;
      }
   print_result(name, iterations, BUFFERSIZE, clocks_used);
   }

template<typename S>
void bench_scipher(const std::string& name, u32bit seconds)
   {
   static const u32bit BUFFERSIZE = OpenCL::DEFAULT_BUFFERSIZE;
   S cipher;
   cipher.set_key(key);
   byte buf[BUFFERSIZE] = { 0 };
   u32bit iterations = 1, j = 0;
   u32bit start = std::clock(), clocks_used = 0;
   while(clocks_used < FACTOR * seconds * CLOCKS_PER_SEC)
      {
      iterations *= 2;
      for(; j != iterations; j++)
         cipher.encrypt(buf, BUFFERSIZE);
      clocks_used = std::clock() - start;
      }
   print_result(name, iterations, BUFFERSIZE, clocks_used);
   }

template<typename R>
void bench_rng(const std::string& name, u32bit seconds)
   {
   static const u32bit BUFFERSIZE = 1;
   R rng;
   u32bit iterations = 1, j = 0;
   u32bit start = std::clock(), clocks_used = 0;
   while(clocks_used < FACTOR * seconds * CLOCKS_PER_SEC)
      {
      iterations *= 2;
      for(; j != iterations; j++)
         rng.random();
      clocks_used = std::clock() - start;
      }
   print_result(name, iterations, BUFFERSIZE, clocks_used);
   }

template<typename H>
void bench_hash(const std::string& name, u32bit seconds)
   {
   static const u32bit BUFFERSIZE = OpenCL::DEFAULT_BUFFERSIZE;
   H hash;
   byte buf[BUFFERSIZE];
   OpenCL::X917<OpenCL::Square> rng;
   rng.randomize(buf, BUFFERSIZE);
   u32bit iterations = 1, j = 0;
   u32bit start = std::clock(), clocks_used = 0;
   while(clocks_used < FACTOR * seconds * CLOCKS_PER_SEC)
      {
      iterations *= 2;
      for(; j != iterations; j++)
         hash.update(buf, BUFFERSIZE);
      clocks_used = std::clock() - start;
      }
   print_result(name, iterations, BUFFERSIZE, clocks_used);
   }

template<typename M>
void bench_mac(const std::string& name, u32bit seconds)
   {
   static const u32bit BUFFERSIZE = OpenCL::DEFAULT_BUFFERSIZE;
   M mac;
   mac.set_key(key);
   byte buf[BUFFERSIZE];
   OpenCL::X917<OpenCL::Square> rng;
   rng.randomize(buf, BUFFERSIZE);
   u32bit iterations = 1, j = 0;
   u32bit start = std::clock(), clocks_used = 0;
   while(clocks_used < FACTOR * seconds * CLOCKS_PER_SEC)
      {
      iterations *= 2;
      for(; j != iterations; j++)
         mac.update(buf, BUFFERSIZE);
      clocks_used = std::clock() - start;
      }
   print_result(name, iterations, BUFFERSIZE, clocks_used);
   }

void benchmark_block_ciphers(u32bit seconds)
   {
   bench_bcipher<OpenCL::Blowfish>("Blowfish", seconds);
   bench_bcipher<OpenCL::CAST256>("CAST256", seconds);
   bench_bcipher<OpenCL::CAST5>("CAST5", seconds);
   bench_bcipher<OpenCL::CS_Cipher>("CS-Cipher", seconds);
   bench_bcipher<OpenCL::DES>("DES", seconds);
   bench_bcipher<OpenCL::DESX>("DESX", seconds);
   bench_bcipher<OpenCL::TripleDES>("Triple-DES", seconds);
   bench_bcipher<OpenCL::GOST>("GOST", seconds);
   bench_bcipher<OpenCL::IDEA>("IDEA", seconds);
   bench_bcipher< OpenCL::LubyRackoff<OpenCL::MD5> >("Luby-Rackoff<MD5>",
                                                      seconds);
   bench_bcipher<OpenCL::MISTY1>("MISTY1", seconds);
   bench_bcipher<OpenCL::RC2>("RC2", seconds);
   bench_bcipher<OpenCL::RC5_12>("RC5 (r = 12)", seconds);
   bench_bcipher<OpenCL::RC5_16>("RC5 (r = 16)", seconds);
   bench_bcipher<OpenCL::RC6>("RC6", seconds);
   bench_bcipher<OpenCL::Rijndael>("Rijndael (r = 14)", seconds);
   bench_bcipher<OpenCL::SAFER_SK128>("SAFER-SK128", seconds);
   bench_bcipher<OpenCL::Serpent>("Serpent", seconds);
   bench_bcipher<OpenCL::SHARK>("SHARK", seconds);
   bench_bcipher<OpenCL::Skipjack>("Skipjack", seconds);
   bench_bcipher<OpenCL::Square>("Square", seconds);
   bench_bcipher<OpenCL::TEA>("TEA", seconds);
   bench_bcipher<OpenCL::ThreeWay>("ThreeWay", seconds);
   bench_bcipher<OpenCL::Twofish>("Twofish", seconds);
   bench_bcipher<OpenCL::XTEA>("XTEA", seconds);
   }

void benchmark_stream_ciphers(u32bit seconds)
   {
   bench_scipher<OpenCL::ARC4>("ARC4", seconds);
   bench_scipher<OpenCL::ISAAC>("ISAAC", seconds);
   bench_scipher<OpenCL::SEAL>("SEAL", seconds);
   }

void benchmark_rngs(u32bit seconds)
   {
   bench_rng<OpenCL::Randpool>("Randpool", seconds);
   bench_rng< OpenCL::X917<OpenCL::Square> >("X917<Square>", seconds);
   }

void benchmark_hash_functions(u32bit seconds)
   {
   bench_hash<OpenCL::Adler32>("Adler32", seconds);
   bench_hash<OpenCL::CRC24>("CRC24", seconds);
   bench_hash<OpenCL::CRC32>("CRC32", seconds);
   bench_hash<OpenCL::HAVAL_256>("HAVAL (PASS = 5)", seconds);
   bench_hash<OpenCL::MD2>("MD2", seconds);
   bench_hash<OpenCL::MD4>("MD4", seconds);
   bench_hash<OpenCL::MD5>("MD5", seconds);
   bench_hash<OpenCL::RIPEMD128>("RIPE-MD128", seconds);
   bench_hash<OpenCL::RIPEMD160>("RIPE-MD160", seconds);
   bench_hash<OpenCL::SHA1>("SHA-1", seconds);
   bench_hash<OpenCL::SHA2_256>("SHA-256", seconds);
   bench_hash<OpenCL::SHA2_512>("SHA-512", seconds);
   bench_hash<OpenCL::Tiger>("Tiger", seconds);
   }

void benchmark_macs(u32bit seconds)
   {
   // Don't bother with HMAC-* as the performance is basically identical
   // to the hash used.
   bench_mac< OpenCL::EMAC<OpenCL::Square> >("EMAC<Square>", seconds);
   bench_mac<OpenCL::MD5MAC>("MD5-MAC", seconds);
   }
