/*
Various utility functions for the validation code
This file is in the public domain
*/

#include <iostream>
#include <string>
#include <typeinfo>

// OpenCL headers
#include "filtbase.h"
#include "symkey.h"

#include "blowfish.h"
#include "cast256.h"
#include "cast5.h"
#include "cscipher.h"
#include "des.h"
#include "gost.h"
#include "idea.h"
#include "lubyrack.h"
#include "misty1.h"
#include "rc2.h"
#include "rc5.h"
#include "rc6.h"
#include "rijndael.h"
#include "safer_sk.h"
#include "serpent.h"
#include "shark.h"
#include "skipjack.h"
#include "square.h"
#include "tea.h"
#include "threeway.h"
#include "twofish.h"
#include "xtea.h"

#include "sha1.h" // for Luby-Rackoff

#include "checksum.h"
using namespace OpenCL;

template<typename B>
class ECB_Filter : public Filter
   {
   public:
      void write(const byte[], u32bit);
      void final();
      ECB_Filter(const BlockCipherKey& key)
        { cipher.set_key(key, key.length()); position = 0; }
   private:
      static const u32bit BLOCKSIZE = B::BLOCKSIZE;
      B cipher;
      SecureBuffer<byte, BLOCKSIZE> buffer;
      u32bit position;
      CRC32 crc_input, crc_decrypt;
   };

template<typename B>
void ECB_Filter<B>::write(const byte input[], u32bit length)
   {
   crc_input.update(input, length);
   buffer.copy(position, input, length);
   if(position + length >= BLOCKSIZE)
      {
      cipher.encrypt(buffer);
      send(buffer, BLOCKSIZE);
      cipher.decrypt(buffer);
      crc_decrypt.update(buffer, BLOCKSIZE);
      input += (BLOCKSIZE - position);
      length -= (BLOCKSIZE - position);
      while(length >= BLOCKSIZE)
         {
         cipher.encrypt(input, buffer);
         send(buffer, BLOCKSIZE);
         cipher.decrypt(buffer);
         crc_decrypt.update(buffer, BLOCKSIZE);
         input += BLOCKSIZE;
         length -= BLOCKSIZE;
         }
       buffer.copy(input, length);
      position = 0;
      }
   position += length;
   }

template<typename B>
void ECB_Filter<B>::final()
   {
   /* We want to check decryption works too, so we just have a CRC32 hash
      of the decrypted output. If it doesn't equal the CRC of the input, we're
      in trouble. It's done internally, so we don't have to export this odd
      iconsistenctey WRT the other functions into the validate code. If it's
      wrong, we just go and die. */
   byte hash_input[4], hash_decrypt[4];
   crc_decrypt.final(hash_decrypt);
   crc_input.final(hash_input);
   for(int j = 0; j != 4; j++)
      if(hash_input[j] != hash_decrypt[j])
         {
         std::cout << "In " << typeid(cipher).name()
                   << " decryption CRC check failed." << std::endl;
         std::exit(1);
         }
   if(position)
      {
      std::cout << "In ECB_Filter::final, position != 0" << std::endl;
      std::exit(1);
      }
   }

Filter* get_block(const std::string& algname, const SymmetricKey& key)
   {
   if(algname == "Blowfish")         return new ECB_Filter<Blowfish>(key);
   else if(algname == "CAST256")     return new ECB_Filter<CAST256>(key);
   else if(algname == "CAST5")       return new ECB_Filter<CAST5>(key);
   else if(algname == "CS-Cipher")   return new ECB_Filter<CS_Cipher>(key);
   else if(algname == "DES")         return new ECB_Filter<DES>(key);
   else if(algname == "DESX")        return new ECB_Filter<DESX>(key);
   else if(algname == "Triple-DES")  return new ECB_Filter<TripleDES>(key);
   else if(algname == "GOST")        return new ECB_Filter<GOST>(key);
   else if(algname == "IDEA")        return new ECB_Filter<IDEA>(key);
   else if(algname == "Luby-Rackoff")
      return new ECB_Filter< LubyRackoff<SHA1> >(key);
   else if(algname == "MISTY1")      return new ECB_Filter<MISTY1>(key);
   else if(algname == "RC2")         return new ECB_Filter<RC2>(key);
   else if(algname == "RC5-12")      return new ECB_Filter<RC5_12>(key);
   else if(algname == "RC5-16")      return new ECB_Filter<RC5_16>(key);
   else if(algname == "RC6")         return new ECB_Filter<RC6>(key);
   else if(algname == "Rijndael")    return new ECB_Filter<Rijndael>(key);
   else if(algname == "SAFER-SK128") return new ECB_Filter<SAFER_SK128>(key);
   else if(algname == "Serpent")     return new ECB_Filter<Serpent>(key);
   else if(algname == "SHARK")       return new ECB_Filter<SHARK>(key);
   else if(algname == "Skipjack")    return new ECB_Filter<Skipjack>(key);
   else if(algname == "Square")      return new ECB_Filter<Square>(key);
   else if(algname == "TEA")         return new ECB_Filter<TEA>(key);
   else if(algname == "ThreeWay")    return new ECB_Filter<ThreeWay>(key);
   else if(algname == "Twofish")     return new ECB_Filter<Twofish>(key);
   else if(algname == "XTEA")        return new ECB_Filter<XTEA>(key);

   else return 0;
   }
