/*
 * Test Driver for OpenCL
 * This file is in the public domain
 */

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
#define WINDOWS_FS
#endif

#include <iostream>
#include <cstdlib>
#include <cstring>
#include <exception>

//#define TEST_TYPES
#if defined(TEST_TYPES)
   #include <limits>
#endif

#include "config.h"
#include "exceptn.h"
using namespace OpenCL_types;

#if defined(WINDOWS_FS)
const std::string VALIDATION_FILE = "checks\\validate.dat";
#else
const std::string VALIDATION_FILE = "checks/validate.dat";
#endif

const u32bit BENCH_SECONDS = 2;

bool produce_html = false;

int validate(const std::string&);
int print_help();

int benchmark(u32bit seconds);
int benchmark_block_ciphers(u32bit seconds);
int benchmark_stream_ciphers(u32bit seconds);
int benchmark_rngs(u32bit seconds);
int benchmark_hash_functions(u32bit seconds);
int benchmark_macs(u32bit seconds);

int main(int argc, char* argv[])
   {
   if(argc != 2 && argc != 3 && argc != 4)
      return print_help();

   for(int j = 1; j != argc; j++)
      if(std::strcmp(argv[j], "--html") == 0)
         produce_html = true;

   for(int j = 1; j != argc; j++)
      {
      if(std::strcmp(argv[j], "--validate") == 0)
         return validate(VALIDATION_FILE);
      if(std::strcmp(argv[j], "--bench-all") == 0)
         return benchmark(BENCH_SECONDS);
      if(std::strcmp(argv[j], "--bench-block") == 0)
         return benchmark_block_ciphers(BENCH_SECONDS);
      if(std::strcmp(argv[j], "--bench-stream") == 0)
         return benchmark_stream_ciphers(BENCH_SECONDS);
      if(std::strcmp(argv[j], "--bench-rng") == 0)
         return benchmark_rngs(BENCH_SECONDS);
      if(std::strcmp(argv[j], "--bench-hash") == 0)
         return benchmark_hash_functions(BENCH_SECONDS);
      if(std::strcmp(argv[j], "--bench-mac") == 0)
         return benchmark_macs(BENCH_SECONDS);
      if(std::strcmp(argv[j], "--html") != 0 &&
         std::strcmp(argv[j], "--help") != 0)
         {
         std::cout << "Unknown option \"" << argv[j] << "\"" << std::endl;
         return 2;
         }
      }

   return print_help();
   }

int print_help()
   {
   std::cout << "Usage:" << std::endl
             << "   --validate: Check test vectors" << std::endl
             << "   --html: Produce HTML output for --bench-all" << std::endl
             << "   --bench-{all,block,stream,rng,hash,mac}: Benchmark"
             << std::endl;
   return 2;
   }

int validate(const std::string& validation_file)
   {
   void test_types();
   u32bit do_validation_tests(const std::string&);

   std::cout << "Beginning validation tests..." << std::endl;
   test_types();
   u32bit errors = 0;
   try
      {
      errors = do_validation_tests(validation_file);
      }
   catch(OpenCL::Exception& e)
      {
      std::cout << "OpenCL exception caught: " << e.what() << std::endl;
      return 1;
      }
   catch(std::exception& e)
      {
      std::cout << "Standard library exception caught: "
                << e.what() << std::endl;
      return 1;
      }
   catch(...)
      {
      std::cout << "Unknown exception caught." << std::endl;
      return 1;
      }
   if(errors == 0)
      std::cout << "All tests passed!" << std::endl;
   else
      std::cout << errors << " test"  << ((errors == 1) ? "" : "s")
                << " failed." << std::endl;
   return (errors ? 1 : 0);
   }

#if defined(TEST_TYPES)
template<typename T>
bool test(const char* type, int digits, bool is_signed)
   {
   bool passed = true;
   if(std::numeric_limits<T>::digits != digits && digits != 0)
      {
      std::cout << "ERROR: numeric_limits<" << type << ">::digits != "
                << digits << std::endl;
      passed = false;
      }
   if(std::numeric_limits<T>::is_signed != is_signed)
      {
      std::cout << "ERROR: numeric_limits<" << type << ">::is_signed != "
                << is_signed << std::endl;
      passed = false;
      }
   if(std::numeric_limits<T>::is_integer == false)
      {
      std::cout << "ERROR: numeric_limits<" << type
                << ">::is_integer == false " << std::endl;
      passed = false;
      }
   return passed;
   }
#endif

void test_types()
   {
   bool passed = true;

#if defined(TEST_TYPES)
   passed = passed && test<byte  >("byte",    8, false);
   passed = passed && test<u16bit>("u16bit", 16, false);
   passed = passed && test<u32bit>("u32bit", 32, false);
   passed = passed && test<u64bit>("u64bit", 64, false);
   passed = passed && test<s32bit>("s32bit", 31,  true);
   passed = passed && test<word  >("word",    0,  true);
   passed = passed && test<dword >("dword",   0,  true);
   if(std::numeric_limits<dword>::digits !=
      2*std::numeric_limits<word>::digits)
      std::cout << "ERROR: numeric_limits<dword>::digits != "
                   "2*numeric_limits<word>::digits" << std::endl;
#endif

   if(!passed)
      {
      std::cout << "Important settings in config.h are wrong. Please fix "
                   "and recompile." << std::endl;
      std::exit(1);
      }
   }

