/*************************************************
* DES Source File                                *
* (C) 1999-2001 The OpenCL Project               *
*************************************************/

#include "des.h"

namespace OpenCL {

/*************************************************
* DES Encryption                                 *
*************************************************/
void DES::encrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   u32bit left  = make_u32bit(in[0], in[1], in[2], in[3]);
   u32bit right = make_u32bit(in[4], in[5], in[6], in[7]);
   IP(left, right);       round(left, right, 0); round(right, left, 1);
   round(left, right, 2); round(right, left, 3); round(left, right, 4);
   round(right, left, 5); round(left, right, 6); round(right, left, 7);
   round(left, right, 8); round(right, left, 9); round(left, right,10);
   round(right, left,11); round(left, right,12); round(right, left,13);
   round(left, right,14); round(right, left,15); FP(left, right);
   out[0] = get_byte(0, right); out[1] = get_byte(1, right);
   out[2] = get_byte(2, right); out[3] = get_byte(3, right);
   out[4] = get_byte(0, left);  out[5] = get_byte(1, left);
   out[6] = get_byte(2, left);  out[7] = get_byte(3, left);
   }

/*************************************************
* DES Decryption                                 *
*************************************************/
void DES::decrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   u32bit left  = make_u32bit(in[0], in[1], in[2], in[3]);
   u32bit right = make_u32bit(in[4], in[5], in[6], in[7]);
   IP(left, right);       round(left, right,15); round(right, left,14);
   round(left, right,13); round(right, left,12); round(left, right,11);
   round(right, left,10); round(left, right, 9); round(right, left, 8);
   round(left, right, 7); round(right, left, 6); round(left, right, 5);
   round(right, left, 4); round(left, right, 3); round(right, left, 2);
   round(left, right, 1); round(right, left, 0); FP(left, right);
   out[0] = get_byte(0, right); out[1] = get_byte(1, right);
   out[2] = get_byte(2, right); out[3] = get_byte(3, right);
   out[4] = get_byte(0, left);  out[5] = get_byte(1, left);
   out[6] = get_byte(2, left);  out[7] = get_byte(3, left);
   }

/*************************************************
* DES Round                                      *
*************************************************/
void DES::round(u32bit& left, u32bit right, u32bit round) const
   {
   u32bit T = rotate_right(right, 4) ^ round_key[2*round];
   left ^= SPBOX1[get_byte(0, T) & 0x3F] ^ SPBOX3[get_byte(1, T) & 0x3F] ^
           SPBOX5[get_byte(2, T) & 0x3F] ^ SPBOX7[get_byte(3, T) & 0x3F];
   T = right ^ round_key[2*round + 1];
   left ^= SPBOX2[get_byte(0, T) & 0x3F] ^ SPBOX4[get_byte(1, T) & 0x3F] ^
           SPBOX6[get_byte(2, T) & 0x3F] ^ SPBOX8[get_byte(3, T) & 0x3F];
   }

/*************************************************
* DES Initial Permutation                        *
*************************************************/
void DES::IP(u32bit& left, u32bit& right) const
   {
   mask(left, right,  4, 0x0F0F0F0F); mask(left, right, 16, 0x0000FFFF);
   mask(right, left,  2, 0x33333333); mask(right, left,  8, 0x00FF00FF);
   right = rotate_left(right, 1);     mask(left, right,  0, 0xAAAAAAAA);
   left  = rotate_left(left, 1);
   }

/*************************************************
* DES Final Permutation                          *
*************************************************/
void DES::FP(u32bit& left, u32bit& right) const
   {
   right = rotate_right(right, 1);    mask(left, right,  0, 0xAAAAAAAA);
   left  = rotate_right(left, 1);     mask(left, right,  8, 0x00FF00FF);
   mask(left, right,  2, 0x33333333); mask(right, left, 16, 0x0000FFFF);
   mask(right, left,  4, 0x0F0F0F0F);
   }

/*************************************************
* Mask Operation for Swapping Bits               *
*************************************************/
void DES::mask(u32bit& left, u32bit& right,
                      byte shift, u32bit mask) const
   {
   u32bit temp = ((left >> shift) ^ right) & mask;
   right ^= temp; left ^= temp << shift;
   }

/*************************************************
* DES Key Setup                                  *
*************************************************/
void DES::set_key(const byte key[], u32bit length) throw(InvalidKeyLength)
   {
   if(length != KEYLENGTH)
      throw InvalidKeyLength("DES");
   static const byte ROT[16] = { 1, 1, 2, 2, 2, 2, 2, 2,
                                 1, 2, 2, 2, 2, 2, 2, 1 };
   static const byte PC1[56] = {
      57, 49, 41, 33, 25, 17,  9,  1, 58, 50, 42, 34, 26, 18, 10,  2,
      59, 51, 43, 35, 27, 19, 11,  3, 60, 52, 44, 36, 63, 55, 47, 39,
      31, 23, 15,  7, 62, 54, 46, 38, 30, 22, 14,  6, 61, 53, 45, 37,
      29, 21, 13,  5, 28, 20, 12,  4 };
   u32bit C = 0, D = 0;
   for(u32bit j = 0; j != 28; j++)
      C = (C << 1) | ((key[(PC1[j]-1)/8] & (1 << (7-((PC1[j]-1)%8)))) ? 1 : 0);
   for(u32bit j = 28; j != 56; j++)
      D = (D << 1) | ((key[(PC1[j]-1)/8] & (1 << (7-((PC1[j]-1)%8)))) ? 1 : 0);
   for(u32bit j = 0; j != 16; j++)
      {
      C = ((C << ROT[j]) | (C >> (28-ROT[j]))) & 0x0FFFFFFF;
      D = ((D << ROT[j]) | (D >> (28-ROT[j]))) & 0x0FFFFFFF;
      round_key[2*j  ] = ((C << 15) & 0x20020000) | ((C << 17) & 0x10000000) |
                         ((C << 10) & 0x08000000) | ((C << 22) & 0x04000000) |
                         ((C >>  2) & 0x02000000) | ((C <<  1) & 0x01000000) |
                         ((C << 16) & 0x00200000) | ((C << 11) & 0x00100000) |
                         ((C <<  3) & 0x00080000) | ((C >>  6) & 0x00040000) |
                         ((C >>  4) & 0x00010000) | ((D >>  2) & 0x00002000) |
                         ((D <<  8) & 0x00001000) | ((D >> 14) & 0x00000808) |
                         ((D >>  9) & 0x00000400) | ((D      ) & 0x00000200) |
                         ((D <<  7) & 0x00000100) | ((D >>  7) & 0x00000020) |
                         ((D >>  3) & 0x00000011) | ((D <<  2) & 0x00000004) |
                         ((D >> 21) & 0x00000002);
      round_key[2*j+1] = ((C <<  4) & 0x24000000) | ((C << 28) & 0x10000000) |
                         ((C << 14) & 0x08000000) | ((C << 18) & 0x02080000) |
                         ((C <<  6) & 0x01000000) | ((C <<  9) & 0x00200000) |
                         ((C >>  1) & 0x00100000) | ((C << 10) & 0x00040000) |
                         ((C <<  2) & 0x00020000) | ((C >> 10) & 0x00010000) |
                         ((D >> 13) & 0x00002000) | ((D >>  4) & 0x00001000) |
                         ((D <<  6) & 0x00000800) | ((D >>  1) & 0x00000400) |
                         ((D >> 14) & 0x00000200) | ((D      ) & 0x00000100) |
                         ((D >>  5) & 0x00000020) | ((D >> 10) & 0x00000010) |
                         ((D >>  3) & 0x00000008) | ((D >> 18) & 0x00000004) |
                         ((D >> 26) & 0x00000002) | ((D >> 24) & 0x00000001);
      }
   }

/*************************************************
* DESX Encryption                                *
*************************************************/
void DESX::encrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   xor_buf(out, in, K1, BLOCKSIZE);
   des.encrypt(out);
   xor_buf(out, K2, BLOCKSIZE);
   }

/*************************************************
* DESX Decryption                                *
*************************************************/
void DESX::decrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   xor_buf(out, in, K2, BLOCKSIZE);
   des.decrypt(out);
   xor_buf(out, K1, BLOCKSIZE);
   }

/*************************************************
* DESX Key Setup                                 *
*************************************************/
void DESX::set_key(const byte key[], u32bit length) throw(InvalidKeyLength)
   {
   if(length != KEYLENGTH)
      throw InvalidKeyLength("DESX");
   K1.copy(key, BLOCKSIZE);
   des.set_key(key + BLOCKSIZE, BLOCKSIZE);
   K2.copy(key + 2 * BLOCKSIZE, BLOCKSIZE);
   }


}

