/*************************************************
* MISTY1 Source File                             *
* (C) 1999-2001 The OpenCL Project               *
*************************************************/

#include "misty1.h"

namespace OpenCL {

/*************************************************
* MISTY1 Encryption                              *
*************************************************/
void MISTY1::encrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   u32bit B0 = make_u32bit(in[0], in[1], in[2], in[3]);
   u32bit B1 = make_u32bit(in[4], in[5], in[6], in[7]);
   FLE(B0, 0); FLO(B1, 1); B1 ^= FO(B0, 0); B0 ^= FO(B1, 1);
   FLE(B0, 2); FLO(B1, 3); B1 ^= FO(B0, 2); B0 ^= FO(B1, 3);
   FLE(B0, 4); FLO(B1, 5); B1 ^= FO(B0, 4); B0 ^= FO(B1, 5);
   FLE(B0, 6); FLO(B1, 7); B1 ^= FO(B0, 6); B0 ^= FO(B1, 7);
   FLE(B0, 8); FLO(B1, 9);
   out[0] = get_byte(0, B1); out[1] = get_byte(1, B1);
   out[2] = get_byte(2, B1); out[3] = get_byte(3, B1);
   out[4] = get_byte(0, B0); out[5] = get_byte(1, B0);
   out[6] = get_byte(2, B0); out[7] = get_byte(3, B0);
   }

/*************************************************
* MISTY1 Decryption                              *
*************************************************/
void MISTY1::decrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   u32bit B0 = make_u32bit(in[4], in[5], in[6], in[7]);
   u32bit B1 = make_u32bit(in[0], in[1], in[2], in[3]);
   FLINVE(B0, 8); FLINVO(B1, 9); B0 ^= FO(B1, 7); B1 ^= FO(B0, 6);
   FLINVE(B0, 6); FLINVO(B1, 7); B0 ^= FO(B1, 5); B1 ^= FO(B0, 4);
   FLINVE(B0, 4); FLINVO(B1, 5); B0 ^= FO(B1, 3); B1 ^= FO(B0, 2);
   FLINVE(B0, 2); FLINVO(B1, 3); B0 ^= FO(B1, 1); B1 ^= FO(B0, 0);
   FLINVE(B0, 0); FLINVO(B1, 1);
   out[0] = get_byte(0, B0); out[1] = get_byte(1, B0);
   out[2] = get_byte(2, B0); out[3] = get_byte(3, B0);
   out[4] = get_byte(0, B1); out[5] = get_byte(1, B1);
   out[6] = get_byte(2, B1); out[7] = get_byte(3, B1);
   }

/*************************************************
* MISTY1 FLE Function                            *
*************************************************/
void MISTY1::FLE(u32bit& input, u32bit index) const
   {
   u16bit T0 = (u16bit)(input >> 16), T1 = (u16bit)(input & 0xFFFF);
   T1 ^= T0 & EK[index/2]; T0 ^= T1 | EK[(index/2+6)%8+8];
   input = ((T0 << 16) | T1);
   }

/*************************************************
* MISTY1 FLO Function                            *
*************************************************/
void MISTY1::FLO(u32bit& input, u32bit index) const
   {
   u16bit T0 = (u16bit)(input >> 16), T1 = (u16bit)(input & 0xFFFF);
   T1 ^= T0 & EK[(index/2 + 2) % 8 + 8]; T0 ^= T1 | EK[(index/2 + 4) % 8];
   input = ((T0 << 16) | T1);
   }

/*************************************************
* MISTY1 FLINVE Function                         *
*************************************************/
void MISTY1::FLINVE(u32bit& input, byte index) const
   {
   u16bit T0 = (u16bit)(input >> 16), T1 = (u16bit)(input & 0xFFFF);
   T0 ^= T1 | EK[(index/2 + 6) % 8 + 8]; T1 ^= T0 & EK[index/2];
   input = ((T0 << 16) | T1);
   }

/*************************************************
* MISTY1 FLINVO Function                         *
*************************************************/
void MISTY1::FLINVO(u32bit& input, byte index) const
   {
   u16bit T0 = (u16bit)(input >> 16), T1 = (u16bit)(input & 0xFFFF);
   T0 ^= T1 | EK[(index/2 + 4) % 8]; T1 ^= T0 & EK[(index/2 + 2) % 8 + 8];
   input = ((T0 << 16) | T1);
   }

/*************************************************
* MISTY1 FO Function                             *
*************************************************/
u32bit MISTY1::FO(u32bit input, u32bit index) const
   {
   u16bit T0 = (u16bit)(input >> 16), T1 = (u16bit)(input & 0xFFFF);
   T0 ^= EK[index]; T0 = FI(T0, EK[(index + 5) % 8 + 8]) ^ T1;
   T1 ^= EK[(index + 2) % 8]; T1 = FI(T1, EK[(index + 1) % 8 + 8]);
   T1 ^= T0; T0 ^= EK[(index + 7) % 8]; T0 = FI(T0, EK[(index + 3) % 8 + 8]);
   T0 ^= T1; T1 ^= EK[(index + 4) % 8]; return ((T1 << 16) | T0);
   }

/*************************************************
* MISTY1 FI Function                             *
*************************************************/
u16bit MISTY1::FI(u16bit input, u16bit key) const
   {
   u16bit D9 = (u16bit)(input >> 7), D7 = (u16bit)(input & 0x7F);
   D9 = S9[D9] ^ D7; D7 = (u16bit)((S7[D7] ^ D9) & 0x7F);
   D7 ^= (u16bit)(key >> 9); D9 ^= (u16bit)(key & 0x1FF);
   D9 = S9[D9] ^ D7; return (u16bit)((D7 << 9) | D9);
   }

/*************************************************
* MISTY1 Key Setup                               *
*************************************************/
void MISTY1::set_key(const byte key[], u32bit length) throw(InvalidKeyLength)
   {
   if(length != KEYLENGTH)
      throw InvalidKeyLength("MISTY1");
   for(u32bit j = 0; j != KEYLENGTH; j++)
      EK[j] = make_u16bit(key[2*j], key[2*j+1]);
   for(u32bit j = 0; j != 8; j++)
      {
      EK[j+ 8] = FI(EK[j], EK[(j+1) % 8]);
      EK[j+16] = (u16bit)(EK[j+8] & 0x1FF);
      EK[j+24] = (u16bit)(EK[j+8] >> 9);
      }
   }

}
